from evaluation_framework.utils.objectIO_utils import load_obj
from evaluation_framework.utils.memmap_utils import write_memmap
from evaluation_framework.utils.memmap_utils import read_memmap

import copy
import numpy as np
import pandas as pd
import os


class TaskGraph():
    """
    The atomic task graph that is run by each dask process. 
    """
    
    def __init__(self, task_manager, cv): 

        self.task_manager = task_manager
        self.cv = cv

    def run(self, group_key, cv_split_index):   

        attempts = 0

        while attempts < 3:

            try:
                
                train_data, test_data, train_idx, test_idx = self.get_data(group_key, cv_split_index)
                prediction_result, evaluation_result = self.task_graph(train_data, test_data, group_key)

                if self.task_manager.return_predictions:
                    self.record_predictions(group_key, cv_split_index, prediction_result, test_data, test_idx)

                break

            except:

                attempts += 1

        train_data, test_data, train_idx, test_idx = self.get_data(group_key, cv_split_index)
        prediction_result, evaluation_result = self.task_graph(train_data, test_data, group_key)

        if self.task_manager.return_predictions:
            self.record_predictions(group_key, cv_split_index, prediction_result, test_data, test_idx)

        return (group_key, cv_split_index, evaluation_result, len(prediction_result))

    def get_data(self, group_key, cv_split_index):

        memmap_root_dirpath = os.path.join(os.getcwd(), self.task_manager.memmap_root_dirname)
        memmap_map_filepath = os.path.join(memmap_root_dirpath, 'memmap_map')
        self.memmap_map = load_obj(memmap_map_filepath)
        
        train_idx, test_idx = self._get_cross_validation_fold_idx(self.memmap_map, group_key, cv_split_index)

        train_data = self._read_memmap(self.memmap_map, group_key, train_idx)
        test_data = self._read_memmap(self.memmap_map, group_key, test_idx)

        return train_data, test_data, train_idx, test_idx

    def task_graph(self, train_data, test_data, group_key):  # groupkey is redundant info get rid of it
        
        configs = self.task_manager.user_configs
        
        preprocessed_train_data = self.task_manager.preprocess_train_data(
            train_data, 
            configs)
        
        trained_estimator = self.task_manager.model_fit(
           preprocessed_train_data, 
           self.task_manager.hyperparameters, 
           self.task_manager.estimator,
           self.task_manager.feature_names[group_key],
           self.task_manager.target_name)

        preprocessed_test_data = self.task_manager.preprocess_test_data(
           test_data, 
           preprocessed_train_data, 
           configs)

        prediction_result = self.task_manager.model_predict(
           preprocessed_test_data, 
           trained_estimator, 
           self.task_manager.feature_names[group_key],
           self.task_manager.target_name)

        evaluation_result = self.task_manager.evaluate_prediction(
           preprocessed_test_data, 
           prediction_result['specialEF_float32_predictions'])

        return (prediction_result, evaluation_result)
        
    def _read_memmap(self, memmap_map, group_key, data_idx):
    
        missing_keys = memmap_map['groups'][group_key]['attributes']['missing_keys']
        data_colnames = copy.copy(memmap_map['groups'][group_key]['attributes']['numeric_keys']) 

        filepath = os.path.join(memmap_map['root_dirpath'], memmap_map['groups'][group_key]['arrays']['numeric_types']['filepath'])
        dtype = memmap_map['groups'][group_key]['arrays']['numeric_types']['dtype']
        shape = memmap_map['groups'][group_key]['arrays']['numeric_types']['shape']
        data_arrays = [read_memmap(filepath, dtype, shape, data_idx)]

        for colname in missing_keys['datetime_types']:
        
            filepath = os.path.join(memmap_map['root_dirpath'], memmap_map['groups'][group_key]['arrays'][colname]['filepath'])
            dtype = memmap_map['groups'][group_key]['arrays'][colname]['dtype']
            shape = memmap_map['groups'][group_key]['arrays'][colname]['shape']
            tmp_array = read_memmap(filepath, dtype, shape, data_idx)

            data_arrays.append(tmp_array.reshape(-1, 1))
            data_colnames.append(colname)
            
        data_array = np.hstack(data_arrays)
        pdf = pd.DataFrame(data_array, columns=data_colnames)
        
        for i in range(len(missing_keys['datetime_types'])):
            pdf.iloc[:, i-1] = pd.to_datetime(pdf.iloc[:, i-1])
            
        for colname in missing_keys['str_types']:
            
            filepath = os.path.join(memmap_map['root_dirpath'], memmap_map['groups'][group_key]['arrays'][colname]['filepath'])
            dtype = memmap_map['groups'][group_key]['arrays'][colname]['dtype']
            shape = memmap_map['groups'][group_key]['arrays'][colname]['shape']
            tmp_array = read_memmap(filepath, dtype, shape, data_idx)

            tmp_array = tmp_array.astype(str)
            pdf[colname] = tmp_array

        return pdf
    
    def _get_cross_validation_fold_idx(self, memmap_map, group_key, cv_split_index):
        
        if self.task_manager.orderby:  # have another parameter to check orderby needs to happen...
            # by cv scheme itself!
            
            # need to add random state

            filepath = os.path.join(memmap_map['root_dirpath'], memmap_map['groups'][group_key]['arrays']['orderby_array']['filepath'])
            dtype = memmap_map['groups'][group_key]['arrays']['orderby_array']['dtype']
            shape = memmap_map['groups'][group_key]['arrays']['orderby_array']['shape']
            group_ordered_array = read_memmap(filepath, dtype, shape)

            for idx, (train, test) in enumerate(self.cv.split(group_ordered_array)):
                if idx == cv_split_index:
                    break
        
        return train, test
    
    def record_predictions(self, group_key, cv_split_index, prediction_result, test_data, test_idx):
        """memmap['groups'][group_key]['groups'][group_key_innder]['arrays'][filepath, dtype, shape]

        """
        test_data_prediction = test_data.merge(prediction_result, on='specialEF_float32_UUID', how='inner')

        predictions_array = test_data_prediction[['specialEF_float32_UUID', 'specialEF_float32_predictions']]
        predictions_array = predictions_array.values.astype(np.float32)

        filename = '__'.join((group_key, str(cv_split_index))) + '.npy'
        filepath = os.path.join(os.getcwd(), self.task_manager.prediction_records_dirname, filename)

        # try:
        np.save(filepath, predictions_array)
        np.load(filepath)
        # except:
        #     pass
            # need to pass some value to indicate failure instead of unavailability!












from dask.distributed import Client, LocalCluster
from dask_yarn import YarnCluster
from evaluation_framework.utils.decorator_utils import yarn_directory_normalizer

import threading
import queue
import socket
import os
import time

def get_host_ip_address():
    """Get the host ip address of the machine where the executor is running. 

    Returns
    -------
    host_ip : String
    """
    try:
        host_name = socket.gethostname()
        host_ip = socket.gethostbyname(host_name)
        return host_ip
    except:
        host_ip = '127.0.0.1'  # playing with fire...
        return host_ip


class MultiThreadTaskQueue(queue.Queue):
    
    def __init__(self, num_threads=1):
        queue.Queue.__init__(self)
        self.num_threads = num_threads
        self.start_threads()
        self.results = []
        
    def put_task(self, task, *args, **kwargs):
        self.put((task, args, kwargs))
        
    def start_threads(self):
        for i in range(self.num_threads):
            t = threading.Thread(target=self.task_in_thread)
            t.setDaemon(True)
            t.start()
            
    def task_in_thread(self):
        while True:
            task, args, kwargs = self.get()
            result = task(*args, **kwargs)
            self.results.append(result)
            self.task_done()

    def get_results(self):
        return self.results


class DualClientFuture():
    
    def __init__(self, local_client_n_workers, local_client_threads_per_worker,
                 yarn_client_n_workers, yarn_client_worker_vcores, yarn_client_worker_memory, verbose=True):
        
        host_ip = get_host_ip_address()
        
        self.local_cluster = LocalCluster(
            n_workers=local_client_n_workers,
            threads_per_worker=local_client_threads_per_worker, 
            processes=True, 
            host=host_ip)
        self.local_client = Client(address=self.local_cluster, timeout='2s') 
        
        self.yarn_cluster = YarnCluster(
            n_workers=yarn_client_n_workers, 
            worker_vcores=yarn_client_worker_vcores, 
            worker_memory=yarn_client_worker_memory,
            environment="python:///usr/bin/python3")
        self.yarn_client = Client(self.yarn_cluster)

        self.wait_container_resource_alloc()
        
        self.local_client_n_workers = local_client_n_workers
        self.yarn_client_n_workers = yarn_client_n_workers
        
        self.task_counter = 0
        self.yarn_client_n_workers = yarn_client_n_workers

        self.verbose = verbose

    def wait_container_resource_alloc(self):

        while True:
            
            waiting_containers = [yarn_container_obj for yarn_container_obj in self.yarn_cluster.workers() 
                                  if str(yarn_container_obj.state)=='WAITING']
            
            if len(waiting_containers)==0:
                break

            time.sleep(1.0)

    def submit(self, func, *args, **kwargs):

        if self.verbose==True:
            print('total n workers: {}'.format(self.local_client_n_workers + self.yarn_client_n_workers))

        remainder = self.task_counter % (self.local_client_n_workers + self.yarn_client_n_workers)
        
        if remainder <= (self.local_client_n_workers-1):

            if self.verbose==True:
                print('remainder: {}, n_local_worker: {}, running on local'.format(remainder, self.local_client_n_workers))

            future = self.local_client.submit(func, *args, **kwargs)
        else:

            if self.verbose==True:
                print('remainder: {}, n_local_worker: {}, running on remote'.format(remainder, self.local_client_n_workers))

            func = yarn_directory_normalizer(func)
            future = self.yarn_client.submit(func, None, *args, **kwargs)
            
        self.task_counter += 1
        
        return future.result()
    
    def get_worker_ip_addresses(self):
        
        while True:
            yarn_container_objects = self.yarn_cluster.workers()
            if len(yarn_container_objects)==self.yarn_client_n_workers:
                break
            time.sleep(0.1)
            
        ip_addrs = set()
        
        for yarn_container_object in yarn_container_objects:
            ip_addrs.add(yarn_container_object.yarn_node_http_address.split('.')[0].replace('-', '.')[3:])
        
        return list(ip_addrs)
    
    def submit_per_node(self, func, *args, **kwargs):
        
        func = yarn_directory_normalizer(func)

        ip_addrs = self.get_worker_ip_addresses()

        futures = list()
        
        for ip_addr in ip_addrs:
            futures.append(self.yarn_client.submit(func, ip_addr, *args, **kwargs, workers=ip_addr))
        
        return self.yarn_client.gather(futures)
    
    def get_dashboard_link(self):
        
        print('local cluster: ', self.local_cluster.dashboard_link)
        print('yarn cluster:  ', self.yarn_cluster.dashboard_link)


class ClientFuture():
    
    def __init__(self, local_client_n_workers, local_client_threads_per_worker):
        
        host_ip = get_host_ip_address()
        self.local_cluster = LocalCluster(n_workers=local_client_n_workers,
                               threads_per_worker=local_client_threads_per_worker, 
                               processes=True, 
                               host=host_ip)
        self.local_client = Client(address=self.local_cluster, timeout='2s') 
        
    def submit(self, func, *args, **kwargs):
        
        future = self.local_client.submit(func, *args, **kwargs)
        return future.result() 
        
    def get_dashboard_link(self):
        
        print('local cluster: ', self.local_cluster.dashboard_link)




import functools
import time
import os

def failed_method_retry(method, max_retries=15):
    """This is a decorator for allowing certain methods to retry itself in cases of
    failures, as many as max_retries times. This decorator can work together with 
    the above two artificial traceback utility methods. 

    Example usage:

        @object_exception_catcher
        @failed_method_retry
        def run_transformer(self, df, hist_df):

        This will first allow the method to be retried, and then after max_retries times
        of retries, the exception catcher decorator will catch the final exception.
    """
    
    @functools.wraps(method)
    def failed_method_retried(*args, **kwargs):
        
        error = None
        
        for i in range(max_retries):
            
            try:
                return method(*args, **kwargs)
            
            except Exception as e:
                error = e
                print('[{}] method failed due to {}'.format(method.__name__, error))
                time.sleep(1.5)
                continue
            
        else:
            raise error
    
    return failed_method_retried


def yarn_directory_normalizer(base_method):
    
    @functools.wraps(base_method)
    def method_modifier(ip_addr, *args, **kwargs):
        """The ip_addr is for [ submit_per_node ] method in Dask Yarn dual client mode. 
        We pass in different arguments so that Dask's work assignment policy does not override
        the desired behavior of allocating the works across different host ip addresses.
        """
        try:
            
            cwd = os.getcwd()
            
            if os.path.split(cwd)[-1].find('container_')==0:
                app_root_dir = os.path.split(cwd)[0:-1][0]
                
                if os.path.split(app_root_dir)[-1].find('application_')==0:
                    os.chdir(app_root_dir)
            
            try:
                return base_method(*args, **kwargs, ip_addr=ip_addr)
            
            except:
                return base_method(*args, **kwargs)
        
        except Exception as e:
            raise type(e)(str(e) + ' from {} method'.format(base_method.__name__))
        
    return method_modifier





import pandas as pd
import numpy as np

def is_date_str_type(ser):
    
    try:
        pd.to_datetime(ser, format='%Y-%m-%d', errors='raise')
        return True
    except:
        return False
    
def is_not_date_str_type(ser):
    
    try:
        pd.to_datetime(ser.iloc[0:10], format='%Y-%m-%d', errors='raise')
        return False
    except:
        return True
        
def is_numeric_type(ser):
    
    if ser.dtype.kind in 'biufc':
        return True
    else:
        return False

def is_datetime_type(ser):
    
    if ser.dtype.kind in 'Mm':
        return True
    else:
        return False
    
def is_float32_type(ser):
    
    if ser.dtype=='float32':
        return True
    else:
        return False

def cast_datetime2int64(ser):
    
    return ser.astype(np.int64)

def cast_int64_2datetime(ser):
    
    return pd.to_datetime(ser)

def cast_numeric2float32(ser):
    
    return ser.astype(np.float32)

def encode_str2bytes(ser):
    
    lens = ser.str.len()
    maxlen = lens.max().astype(int)
    dtype = 'S{}'.format(maxlen, 1)
    return ser.values.astype(dtype)

def decode_bytes2str(ndarray):

    return ndarray.astype(str)

def encode_date_sequence(ser):
    
    return (ser - ser.min()).dt.days




    









def default_preprocess_train_data(train_data, configs):
    preprocessed_train_data = train_data
    return preprocessed_train_data


def default_preprocess_test_data(test_data, preprocessed_train_data, configs):
    preprocessed_test_data = test_data
    return preprocessed_test_data


def default_evaluate_prediction(preprocessed_test_data, prediction_result):
    
    return prediction_result.mean()







#######################################################################
#######################################################################
## CAUTION : Although you can, try not to override the below methods ##
#######################################################################
#######################################################################
# If you have an unusual way of fitting the model or predicting using a fitted model,
# please specify them in the Estimator class's fit() and predict() method, rather
# than overriding the methods below. 

def default_model_fit(preprocessed_train_data, hyperparameters, estimator, feature_names, target_name):
    
    X = preprocessed_train_data[feature_names]  # need optimization
    
    if False:
        pass
        # read from memmap
    else:
        y = preprocessed_train_data[target_name]
        
    estimator.fit(X, y, hyperparameters)
    
    return estimator
    
def default_model_predict(preprocessed_test_data, trained_estimator, feature_names, target_name):
    
    preprocessed_test_data = preprocessed_test_data.reset_index(drop=True)
    
    X = preprocessed_test_data[feature_names]
    print(X)
    preprocessed_test_data['specialEF_float32_predictions'] = trained_estimator.predict(X)

    prediction_result = preprocessed_test_data[['specialEF_float32_UUID', 'specialEF_float32_predictions']]    
    return prediction_result






__version__ = "0.0.b30"



from ._evaluation_manager.config_setter import ConfigSetter
from ._evaluation_manager.method_setter import MethodSetter

from .evaluation_manager import EvaluationManager

from ._evaluation_engine.data_loader import load_local_data
from ._evaluation_engine.data_loader import upload_local_data
from ._evaluation_engine.data_loader import download_local_data
from ._evaluation_engine.data_loader import upload_remote_data
from ._evaluation_engine.data_loader import download_remote_data

from ._evaluation_engine.dask_futures import MultiThreadTaskQueue
from ._evaluation_engine.dask_futures import ClientFuture
from ._evaluation_engine.dask_futures import DualClientFuture
from ._evaluation_engine.cross_validation_split import DateRollingWindowSplit
from ._evaluation_engine.cross_validation_split import get_cv_splitter

from .task_graph import TaskGraph

from evaluation_framework.utils.objectIO_utils import save_obj
from evaluation_framework.utils.objectIO_utils import load_obj

from evaluation_framework.utils.memmap_utils import write_memmap
from evaluation_framework.utils.memmap_utils import read_memmap

from .evaluation_engine import EvaluationEngine

__all__ = [
	"EvaluationManager",
	"EvaluationEngine",
	]




from future.standard_library import install_aliases
install_aliases()
from urllib.parse import urlparse, urlencode
import os
import boto3
import errno
import shutil


def s3_upload_object(file_path, s3_url, object_name):
	"""Upload a file to S3 bucket in designated key name. 

	Parameters
	----------
	file_path : String
		The path to the file, including the name of the file.
	s3_url : String
		The s3 url up to and including the bucket name.
	object_name : String
		The desired key name to be uploaded in S3 bucket.
	"""
	
	scheme, bucket, path, _, _, _ = urlparse(s3_url)
	key = os.path.join(path[1:], object_name)
	
	s3_resource = boto3.resource('s3')
	s3_client = s3_resource.meta.client
	s3_client.upload_file(Filename=file_path, Bucket=bucket, Key=key)


def s3_download_object(dest_dir, s3_url, object_name_prefix):
	"""Download all objects with designated name prefix from S3 bucket into desired destination directory. 
	If the destination directory does not exist, create one before downloading the objects into it. 

	Parameters
	----------
	dest_dir : String
		The path to the directory into which the objects are to be stored.
	s3_url : String
		The s3 url up to and including the bucket name.
	object_name_prefix : String
		The prefix of the object name to use to filter the download.
	"""
	
	try:
		os.makedirs(dest_dir)
	except OSError as e:
		if e.errno != errno.EEXIST:
			raise
	
	scheme, bucket, path, _, _, _ = urlparse(s3_url)
	prefix = os.path.join(path[1:], object_name_prefix)
	
	s3_resource = boto3.resource('s3')
	s3_bucket = s3_resource.Bucket(name=bucket)
	
	for obj in s3_bucket.objects.filter(Prefix=prefix):
		
		filename = obj.key.split('/')[-1]
		filepath = os.path.join(dest_dir, filename)
		
		s3_bucket.download_file(obj.key, filepath)


def s3_upload_zip_dir(source_dir, s3_url, object_name):
	"""First create a zipped version of the desired directory to upload, and then upload the zipped 
	directory into S3. Delete the zipped directory afterwards. 

	Parameters
	----------
	source_dir : String
		The path of the directory to be zipped and uploaded.
	s3_url : String
		The s3 url up to and including the bucket name.
	object_name : String
		The desired name for the S3 key of the uploaded zipped directory file.
	"""
	
	scheme, bucket, path, _, _, _ = urlparse(s3_url)
	key = os.path.join(path[1:], object_name)
	
	zip_file_dir = shutil.make_archive(source_dir, 'zip', source_dir)
	
	s3_resource = boto3.resource('s3')
	s3_client = s3_resource.meta.client
	s3_client.upload_file(Filename=zip_file_dir, Bucket=bucket, Key=key)
	
	os.remove(zip_file_dir)
	

def s3_delete_object(s3_url, object_name_prefix):
	"""Delete all objects with the designated key name prefix from S3 bucket.
	
	Parameters
	----------
	s3_url : String
		The s3 url up to and including the bucket name.
	object_name_prefix : String
		The key name prefix to filter the deletion. 
	"""
	
	scheme, bucket, path, _, _, _ = urlparse(s3_url)
	prefix = os.path.join(path[1:], object_name_prefix)
	
	s3_resource = boto3.resource('s3')
	s3_bucket = s3_resource.Bucket(name=bucket)
	
	for obj in s3_bucket.objects.filter(Prefix=prefix):
		
		s3_bucket.Object(key=obj.key).delete()


def s3_object_exists(s3_url, object_name):
	"""Return boolean value of True if the object exists in S3, False otherwise.
	
	Parameters
	----------
	s3_url : String
		The s3 url up to and including the bucket name.
	object_name : String
		The total and exact name of the key of the object. 
	"""
	scheme, bucket, path, _, _, _ = urlparse(s3_url)
	key = os.path.join(path[1:], object_name)
	
	s3_resource = boto3.resource('s3')
	
	try:
		s3_resource.Object(bucket_name=bucket, key=key).load()
		return True
		
	except botocore.exceptions.ClientError as e:
		if e.response['Error']['Code'] == "404": 
			return False
		else:
			return False  # in case we want to handle non 404 differently in the future
		

def s3_fetch_keys(s3_url, object_name_prefix):
	"""Get all keys from S3 bucket with the given key name prefix.
	
	Parameters
	----------
	s3_url : String
		The s3 url up to and including the bucket name.
	object_name_prefix : String
		The prefix of the object name to use to filter the search.
	"""
	scheme, bucket, path, _, _, _ = urlparse(s3_url)
	prefix = os.path.join(path[1:], object_name_prefix)
	
	s3_resource = boto3.resource('s3')
	s3_client = s3_resource.meta.client
	
	kwargs = {'Bucket': bucket, 'Prefix': prefix}
	keys = []
	
	while True:
		
		resp = s3_client.list_objects_v2(**kwargs)

		if not 'Contents' in resp:
			break
		
		for obj_metadata_dict in resp['Contents']:
			
			key = obj_metadata_dict['Key']
			keys.append(key)
			
		try:
			kwargs['ContinuationToken'] = resp['NextContinuationToken']
		except KeyError:
			break
	
	return keys


if __name__ == '__main__':

	file_dir = modelfile
	s3_url = 's3://grubhub-gdp-data-transfer-dev/pickup_feasibility'
	object_name = 'model_name_XXX'

	s3_upload_object(file_dir, s3_url, object_name)


	dest_dir = os.getcwd() + '/model_dir_date-XXX'
	s3_url = 's3://grubhub-gdp-data-transfer-dev/pickup_feasibility'
	object_name_prefix = 'model_name'

	s3_download_object(dest_dir, s3_url, object_name_prefix)


	source_dir = dest_dir
	s3_url = 's3://grubhub-gdp-data-transfer-dev/pickup_feasibility'
	object_name = 'zipped_models'

	s3_upload_zip(source_dir, s3_url, object_name)


	s3_url = 's3://grubhub-gdp-data-transfer-dev/pickup_feasibility'
	object_name_prefix = 'model_name'

	s3_delete_object(s3_url, object_name_prefix)



from evaluation_framework.utils.pandas_utils import cast_datetime2int64
from evaluation_framework.utils.pandas_utils import cast_int64_2datetime
from evaluation_framework.utils.pandas_utils import encode_str2bytes
from evaluation_framework.utils.pandas_utils import encode_date_sequence
from evaluation_framework.utils.s3_utils import s3_upload_object
from evaluation_framework.utils.s3_utils import s3_download_object
from evaluation_framework.utils.s3_utils import s3_upload_zip_dir
from evaluation_framework.utils.s3_utils import s3_delete_object
from evaluation_framework.utils.zip_utils import unzip_dir
from evaluation_framework.utils.objectIO_utils import save_obj
from evaluation_framework.utils.objectIO_utils import load_obj
from evaluation_framework.utils.memmap_utils import write_memmap
from evaluation_framework.utils.memmap_utils import read_memmap
from evaluation_framework.utils.decorator_utils import failed_method_retry

import os
import shutil
from collections import namedtuple
import pickle
import numpy as np


@failed_method_retry
def load_local_data(evaluation_manager):

    memmap_root_dirpath = os.path.join(os.getcwd(), evaluation_manager.memmap_root_dirname)

    try:
        os.makedirs(memmap_root_dirpath)
    except:
        shutil.rmtree(memmap_root_dirpath)
        os.makedirs(memmap_root_dirpath)

    if evaluation_manager.return_predictions:

        prediction_records_dirpath = os.path.join(os.getcwd(), evaluation_manager.prediction_records_dirname)

        try:
            os.makedirs(prediction_records_dirpath)
        except:
            shutil.rmtree(prediction_records_dirpath)
            os.makedirs(prediction_records_dirpath)

    memmap_map = _write_memmap_filesys(evaluation_manager, memmap_root_dirpath)
    return memmap_map

def upload_local_data(task_manager):

    memmap_root_dirpath = os.path.join(os.getcwd(), task_manager.memmap_root_dirpath)
    s3_url = task_manager.S3_path
    object_name = task_manager.memmap_root_S3_object_name + '.zip'
    s3_upload_zip_dir(memmap_root_dirpath, s3_url, object_name)

def download_local_data(task_manager):
    """
    1. create memmap dir
    3. translate hdf5 to memmap
    4. graph will just use the memmap dirname to read it off from the "current pos"

    """
    s3_download_object(os.getcwd(), task_manager.S3_path, task_manager.memmap_root_S3_object_name + '.zip')

    zipped_filepath = os.path.join(os.getcwd(), task_manager.memmap_root_S3_object_name + '.zip')
    unzip_dir(zipped_filepath, task_manager.memmap_root_dirname)

    # update memmap_map with new root_dir
    updated_memmap_map_root_dirpath = os.path.join(os.getcwd(), task_manager.memmap_root_dirname)
    memmap_map_filepath = os.path.join(updated_memmap_map_root_dirpath, 'memmap_map')
    memmap_map = load_obj(memmap_map_filepath)
    memmap_map['root_dirpath'] = updated_memmap_map_root_dirpath
    save_obj(memmap_map, memmap_map_filepath)

    if task_manager.return_predictions:

        prediction_records_dirpath = os.path.join(os.getcwd(), task_manager.prediction_records_dirname)

        try:
            os.makedirs(prediction_records_dirpath)
        except:
            shutil.rmtree(prediction_records_dirpath)
            os.makedirs(prediction_records_dirpath)

def upload_remote_data(task_manager, ip_addr):
    """
    1. zip the prediction array directory
    2. send them to S3 bucket

    **The file structure should be identical to that of local machine, making it possible to
    use this method on local machine as well for testing purposes.
    """
    source_dirpath = os.path.join(os.getcwd(), 'prediction_arrays')
    
    job_uuid = task_manager.memmap_root_S3_object_name.split('__')[-1]
    host_uuid = ip_addr.replace('.', '-')
    object_name = 'prediction_arrays' + '__' + host_uuid + '__' + job_uuid + '.zip'
    
    s3_url = task_manager.S3_path
    s3_upload_zip_dir(source_dirpath, s3_url, object_name)
    
    return object_name

def download_remote_data(task_manager):
    """
    1. download the prediction array zip dirs from S3
    2. unzip them and place them into the same directory
    """
    s3_download_object(os.getcwd(), task_manager.S3_path, 'prediction_arrays')
    tmp = os.listdir(os.getcwd())
    prediction_arrays_zips = [elem for elem in tmp if elem.startswith('prediction_array') & elem.endswith('zip')]
    
    for prediction_arrays_zip in prediction_arrays_zips:
        
        zipped_filepath = os.path.join(os.getcwd(), prediction_arrays_zip)
        unzip_dir(zipped_filepath, os.path.join(os.getcwd(), 'prediction_arrays'))
        
def _write_memmap_filesys(task_manager, root_dirpath):
    """memmap mimicking hdf5 filesystem. 
    root_dirpath/
        memmap_map
        groupA__groupA'__arrayA (array)
        groupA__groupA'__arrayB (array)  
        ... etc


    root_dirpath / group_dirpath / filepath
    memmap['groups'][group_key]['groups'][group_key_innder]['arrays'][filepath, dtype, shape]

    """
    memmap_map = dict()
    memmap_map['attributes'] = dict()
    memmap_map['groups'] = dict()  # i.e. we define the first level to be groups
    memmap_map['root_dirpath'] = root_dirpath

    memmap_map_filepath = os.path.join(root_dirpath, 'memmap_map')

    group_key_size_tuples = []

    for group_key, grouped_pdf in task_manager.data.groupby(by=task_manager.groupby):

        group_key_size_tuples.append([group_key, len(grouped_pdf)])

        if task_manager.orderby:
            grouped_pdf = grouped_pdf.sort_values(by=task_manager.orderby)
        grouped_pdf = grouped_pdf.reset_index(drop=True)

        memmap_map['groups'][group_key] = dict()
        source_dirpath = memmap_map['root_dirpath']
        # memmap_map['groups'][group_key]['group_dirpath'] = '__'.join((source_dirpath, group_key))

        # memmap_map['groups'][group_key]['group_dirpath'] = os.path.join(source_dirpath, group_key)  
        memmap_map['groups'][group_key]['group_dirpath'] = group_key

        # NOTE FOR HMF: the first level after root_dirpath needs to be os path join not '__'
        # NOTE FOR HMF: need a check for this for generalized self-describing memmap tool
        # NOTE FOR HMF FROM DUAL CLIENT PERSPECTIVE: we want to leave the root_dirpath flexible
        # so for the first group level, we need to start a new path so we can join with an updated root_dirpath later
        memmap_map['groups'][group_key]['attributes'] = dict()
        memmap_map['groups'][group_key]['arrays'] = dict()  # i.e. we define the second level to be arrays
        
        # NOTE FOR HMF: later, develop this into recursive function and a generalized open source tool

        _write_datetime_types(task_manager, memmap_map, group_key, grouped_pdf)
        _write_str_types(task_manager, memmap_map, group_key, grouped_pdf)
        _write_numeric_types(task_manager, memmap_map, group_key, grouped_pdf)

        if task_manager.orderby:
            _write_orderby_array(task_manager, memmap_map, group_key, grouped_pdf)

        memmap_map['groups'][group_key]['attributes']['numeric_keys'] = task_manager.numeric_types
        memmap_map['groups'][group_key]['attributes']['missing_keys'] = task_manager.missing_keys

    group_key_size_tuples = sorted(group_key_size_tuples, key=lambda x: x[1])
    sorted_group_keys = [elem[0] for elem in group_key_size_tuples]
    memmap_map['attributes']['sorted_group_keys'] = sorted_group_keys

    save_obj(memmap_map, memmap_map_filepath)

    return memmap_map

def _write_datetime_types(task_manager, memmap_map, group_key, grouped_pdf):

    for key in task_manager.missing_keys['datetime_types']:

        # memmap array info
        array = cast_datetime2int64(grouped_pdf[key]).values
        dtype = str(array.dtype)
        shape = array.shape

        memmap_map['groups'][group_key]['arrays'][key] = dict()  # each array object is a dict as well for memmap case, unlike self documenting hdf5

        source_dirpath = memmap_map['groups'][group_key]['group_dirpath']
        filepath = '__'.join((source_dirpath, key))
        memmap_map['groups'][group_key]['arrays'][key]['filepath'] = filepath
        memmap_map['groups'][group_key]['arrays'][key]['dtype'] = dtype
        memmap_map['groups'][group_key]['arrays'][key]['shape'] = shape

        # later, develop this into recursive function and a generalized open source tool

        write_memmap(
            os.path.join(
                memmap_map['root_dirpath'], 
                memmap_map['groups'][group_key]['arrays'][key]['filepath']), 
            memmap_map['groups'][group_key]['arrays'][key]['dtype'], 
            memmap_map['groups'][group_key]['arrays'][key]['shape'], 
            array)

def _write_str_types(task_manager, memmap_map, group_key, grouped_pdf):

    for key in task_manager.missing_keys['str_types']:

        array = encode_str2bytes(grouped_pdf[key])
        dtype = str(array.dtype)
        shape = array.shape

        memmap_map['groups'][group_key]['arrays'][key] = dict()  # each array object is a dict as well for memmap case, unlike self documenting hdf5

        source_dirpath = memmap_map['groups'][group_key]['group_dirpath']
        filepath = '__'.join((source_dirpath, key))
        memmap_map['groups'][group_key]['arrays'][key]['filepath'] = filepath
        memmap_map['groups'][group_key]['arrays'][key]['dtype'] = dtype
        memmap_map['groups'][group_key]['arrays'][key]['shape'] = shape

        write_memmap(
            os.path.join(
                memmap_map['root_dirpath'], 
                memmap_map['groups'][group_key]['arrays'][key]['filepath']), 
            memmap_map['groups'][group_key]['arrays'][key]['dtype'], 
            memmap_map['groups'][group_key]['arrays'][key]['shape'], 
            array)

def _write_numeric_types(task_manager, memmap_map, group_key, grouped_pdf):

    key = 'numeric_types'

    # the code below can be made uniform
    # the below part corresponds to pytable's create_array(group_object, key, array) method
    # can be made into a generalized tool

    array = grouped_pdf[task_manager.numeric_types].values
    dtype = str(array.dtype)
    shape = array.shape

    memmap_map['groups'][group_key]['arrays'][key] = dict()

    source_dirpath = memmap_map['groups'][group_key]['group_dirpath']
    filepath = '__'.join((source_dirpath, key))
    memmap_map['groups'][group_key]['arrays'][key]['filepath'] = filepath
    memmap_map['groups'][group_key]['arrays'][key]['dtype'] = dtype
    memmap_map['groups'][group_key]['arrays'][key]['shape'] = shape

    write_memmap(
        os.path.join(
            memmap_map['root_dirpath'], 
            memmap_map['groups'][group_key]['arrays'][key]['filepath']), 
        memmap_map['groups'][group_key]['arrays'][key]['dtype'], 
        memmap_map['groups'][group_key]['arrays'][key]['shape'], 
        array)

def _write_orderby_array(task_manager, memmap_map, group_key, grouped_pdf):

    key = 'orderby_array'

    array = encode_date_sequence(grouped_pdf[task_manager.orderby]).values
    dtype = str(array.dtype)
    shape = array.shape

    memmap_map['groups'][group_key]['arrays'][key] = dict()

    source_dirpath = memmap_map['groups'][group_key]['group_dirpath']
    filepath = '__'.join((source_dirpath, key))
    memmap_map['groups'][group_key]['arrays'][key]['filepath'] = filepath
    memmap_map['groups'][group_key]['arrays'][key]['dtype'] = dtype
    memmap_map['groups'][group_key]['arrays'][key]['shape'] = shape

    write_memmap(
        os.path.join(
            memmap_map['root_dirpath'], 
            memmap_map['groups'][group_key]['arrays'][key]['filepath']), 
        memmap_map['groups'][group_key]['arrays'][key]['dtype'], 
        memmap_map['groups'][group_key]['arrays'][key]['shape'], 
        array)




from ._evaluation_manager.config_setter import ConfigSetter 
from ._evaluation_manager.method_setter import MethodSetter


class EvaluationManager():
    
    def __init__(self):
        
        self.config_setter = ConfigSetter()
        self.method_setter = MethodSetter()
        
    def setup_evaluation(self, **kwargs):

        configs_set = self.config_setter.set_configs(**kwargs)
        methods_set = self.method_setter.set_methods(config_setter=self.config_setter, **kwargs)

        # if not self.config_setter.set_configs(**kwargs):
        #    return

        # if not self.method_setter.set_methods(config_setter=self.config_setter, **kwargs):
        #    return

        self.load_object_fields(self.config_setter)
        self.load_object_fields(self.method_setter)
        # kill objects
        
    def load_object_fields(self, source_obj):
        
        for k, v in source_obj.__dict__.items():
            self.__dict__[k] = v

            


from ..utils.pandas_utils import is_date_str_type
from ..utils.pandas_utils import is_not_date_str_type
from ..utils.pandas_utils import is_numeric_type
from ..utils.pandas_utils import is_datetime_type
from ..utils.pandas_utils import is_float32_type
from ..utils.pandas_utils import cast_datetime2int64
from ..utils.pandas_utils import cast_numeric2float32
from ..utils.data_structure_utils import get_merged_list_from_dict_list_values
from ..utils.data_structure_utils import dict_is_nested
from ..utils.datetime_utils import check_date_format

import inspect
import pandas as pd
import numpy as np
import os
import warnings
import shutil
import datetime


ORDERED_CV_SCHEMES = ['date_rolling_window']
CV_OPTIONAL_ARGUMENTS = ['orderby', 'train_window', 'test_window']
OPTIONAL_ARGUMENTS = ['groupby', 'hyperparameters', 'user_configs', 'S3_path', 'user_configs', 'return_predictions']
CV_SCHEME_OPTIONS = ['date_rolling_window', 'k_fold', 'binary_classification']
REQUIRED_ESTIMATOR_MEMBER_METHODS = ['fit', 'predict']
FIT_METHOD_PARAMETERS_PARAMETER_NAME = 'parameters'


class ConfigSetter():
    
    def __init__(self):
        
        self.dummy_region_uuid = 'dummy_region'

        self.str_types = []
        self.datetime_types = []
        self.numeric_types = []
        self.original_colnames = []
        
    def set_configs(self, *, estimator=None,
                    data=None, target_name=None, feature_names=None, 
                    hyperparameters=None, cross_validation_scheme=None,
                    groupby=None, 
                    orderby=None, train_window=None, test_window=None,
                    user_configs=None, local_directory_path=None, S3_path=None, 
                    return_predictions=None, **kwargs):
        
        self.estimator = estimator
        self.data = data
        self.target_name = target_name
        self.feature_names = feature_names
        self.hyperparameters = hyperparameters
        self.cross_validation_scheme = cross_validation_scheme
        self.groupby = groupby
        self.orderby = orderby
        self.train_window = train_window
        self.test_window = test_window 
        self.user_configs = user_configs
        self.local_directory_path = local_directory_path
        self.S3_path = S3_path
        self.return_predictions = return_predictions

        print("\u2714 Checking configs requirements...    ", end="", flush=True)

        if not self.passed_arguments_requirements():
            return False
        else:
            print("Passed!")

        print("\u2714 Checking configs validity...        ", end="", flush=True)

        if not self.passed_arguments_validity():
            return False
        else:
            print("Passed!")

        return True
        
    def passed_arguments_requirements(self):
        
        self.required_args = self._required_arguments()
        self.ordered_CV_required_args = self._ordered_CV_required_arguments()
        self.optional_args = self._optional_arguments()

        if len(self.required_args)>0 or len(self.ordered_CV_required_args)>0:
            print('Failed!')
            self._tell_required_arguments()
            return False
        else:
            return True

    def _tell_required_arguments(self):

        if len(self.required_args)>0:
            print('\nRequired config argument(s):\n\n\u25BA {}'.format('  '.join(self.required_args)))
        
        if 'cross_validation_scheme' in self.required_args:
            print('\nAvailable [ cross_validation_scheme ] options:\n\n'
                  '\u25BA {}'.format('  '.join(CV_SCHEME_OPTIONS)))
            
        elif self.cross_validation_scheme in ORDERED_CV_SCHEMES:
            if len(self.ordered_CV_required_args)>0:
                print('\nRequired argument(s) for [ {} ] scheme:\n\n'
                      '\u25BA {}'.format(self.cross_validation_scheme, 
                                         '  '.join(self.ordered_CV_required_args)))

        if len(self.required_args)>0 or len(self.ordered_CV_required_args)>0:
            if len(self.optional_args)>0:
                print('\nOptional config argument(s):\n\n\u25BA {}'.format('  '.join(self.optional_args)))
                print()

    def _required_arguments(self):
        
        required_args = []
        for k, v in self.__dict__.items():
            if v is None and k not in OPTIONAL_ARGUMENTS + CV_OPTIONAL_ARGUMENTS:
                required_args.append(k)
                
        return required_args
    
    def _ordered_CV_required_arguments(self):
        
        ordered_CV_required_args = []
        
        if self.cross_validation_scheme is None:
            return ordered_CV_required_args
        
        elif self.cross_validation_scheme in ORDERED_CV_SCHEMES:
            
            for k, v in self.__dict__.items():
                
                if v is None and k in CV_OPTIONAL_ARGUMENTS:
                    
                    ordered_CV_required_args.append(k)
                
        return ordered_CV_required_args
        
    def _optional_arguments(self):
        
        args_present = []
        
        for k, v in self.__dict__.items():
            if v is not None:
                args_present.append(k)
        
        cv_optional_args = list(set(CV_OPTIONAL_ARGUMENTS) - set(self._ordered_CV_required_arguments()))
        optional_args = cv_optional_args + OPTIONAL_ARGUMENTS
        optional_args = set(optional_args) - set(args_present)     
        return optional_args

    def passed_arguments_validity(self):
        
        self._validate_user_configs()
        self._validate_local_directory_path()
        self._validate_s3_path()
        self._validate_estimator()
        self._validate_target_name()
        self._validate_feature_names()
        self._validate_data()
        self._validate_hyperparameters()
        self._validate_return_predictions()
        return True

    def _validate_return_predictions(self):

        self.prediction_records_dirname = None

        if self.return_predictions is None:
            self.return_predictions = False

        else:

            if not isinstance(self.return_predictions, bool):
                raise TypeError('[ return_predictions ] must be boolean but instead got '
                                '{}'.format(type(self.return_predictions)))

            if self.return_predictions:

                self.prediction_records_dirname = 'prediction_arrays'

                # makes sense to join the predictions on self.data since that data is already
                # loaded in memory. It would be expensive to re-load it into memory, esp. if it
                # is really big.

                # create key column to join the predictions
                key_column = np.arange(len(self.data)).astype(np.float32)
                self.data['specialEF_float32_UUID'] = key_column

                # assuming data validation was done before...
                self.numeric_types.append('specialEF_float32_UUID')

    def _validate_user_configs(self):

        if self.user_configs is None:

            self.user_configs = dict()

    def _validate_local_directory_path(self):

        self.overwrite = True

        if os.path.exists(self.local_directory_path) and not self.overwrite:
            raise ValueError('[ local_dirpath ] already exists. Set [ overwrite ] flag to True '
                             'to use this directory path.')
            
        if os.path.exists(self.local_directory_path):
            shutil.rmtree(self.local_directory_path)
            
        self.initial_dirpath = os.getcwd()
        
        self.memmap_root_dirname = 'memmap_root_dir'
        self.memmap_root_dirpath = os.path.join(self.local_directory_path, self.memmap_root_dirname)

    def _validate_s3_path(self):

        current_time = str(datetime.datetime.now()).replace(" ", '-')
        self.memmap_root_S3_object_name = self.memmap_root_dirname +'__'+ str(datetime.datetime.now()).replace(" ", '-')
        
    def _validate_estimator(self):
        
        member_methods = inspect.getmembers(self.estimator, predicate=inspect.ismethod)
        member_methods = [elem[0] for elem in member_methods]
        
        if not set(REQUIRED_ESTIMATOR_MEMBER_METHODS) < set(member_methods):
            missing_member_methods = set(REQUIRED_ESTIMATOR_MEMBER_METHODS) - set(member_methods)
            print('Failed!')
            raise ValueError('[ estimator ] object is missing \"{}\" '
                             'method(s).'.format(', '.join(missing_member_methods)))
            
        fit_method_parameters = inspect.signature(self.estimator.fit).parameters.keys()
        
        if FIT_METHOD_PARAMETERS_PARAMETER_NAME in fit_method_parameters:
            
            if self.hyperparameters is None:
                print('Failed!')
                raise ValueError('[ estimator ] object requires [ hyperparameters ] argument.')

        else:

            if self.hyperparameters is not None:
                warnings.warn("[ hyperparameters ] is not being used by [ estimator ].")
        
    def _validate_target_name(self):
        
        if self.target_name not in self.data.columns:
            print('Failed!')
            raise ValueError('[ target_name ] "{}" is not in the dataframe.'.format(self.target_name))
            
        if not is_float32_type(self.data[self.target_name]):
            print('Failed!')
            raise TypeError('[ target_name ] "{}" is not of type np.float32.'.format(self.target_name))
            
    def _validate_feature_names(self):
            
        if isinstance(self.feature_names, dict):
            self._check_groupby_presence('feature_names')
            self._check_groupby_contents('feature_names')
        
        # make it into dict if it is not
        else:
            if self.groupby:
                group_keys = self.data[self.groupby].unique()
                feature_names = self.feature_names
                self.feature_names = {group_key: feature_names for group_key in group_keys}
            else:
                self.feature_names = {self.dummy_region_uuid: self.feature_names}
            
        feature_names_list = get_merged_list_from_dict_list_values(self.feature_names)
        
        for feature_name in feature_names_list:
            if not is_float32_type(self.data[feature_name]):
                print('Failed!')
                raise TypeError('[ feature_names ] "{}" is not of type np.float32,'.format(feature_name))
            
        for k, v in self.feature_names.items():
            
            # check the feature names in dataframe
            if not set(v) < set(self.data.columns):
                
                unrecognized_feature_names = set(v) - set(self.data.columns)
                
                if k == self.dummy_region_uuid:
                    print('Failed!')
                    raise ValueError("[ feature_names ] \"{}\" are missing from the input "
                                     "dataframe.".format(', '.join(unrecognized_feature_names)))
                else:
                    print('Failed!')
                    raise ValueError("[ feature_names ] \"{}\" are missing from the input "
                                     "dataframe at group \"{}\".".format(', '.join(unrecognized_feature_names), k))
      
    def _validate_data(self):
        
        if not isinstance(self.data, pd.DataFrame):
            print('Failed!')
            raise TypeError('[ data ] must be of type <pandas.DataFrame>, '
                            'instead got {}.'.format(type(self.data)))
        
        if len(self.data)==0:
            print('Failed!')
            raise ValueError('[ data ] is empty dataframe.')
            
        self.str_types = []
        self.date_str_types = []
        self.datetime_types = []
        self.numeric_types = []
        self.unsupported_types = []
        
        types = self.data.dtypes

        for k, v in types.items():

            if v=='object' and is_not_date_str_type(self.data[k]):
                self.str_types.append(k)
            elif v=='object' and is_date_str_type(self.data[k]):
                self.date_str_types.append(k)
            elif v=='object':
                # raise warning!
                self.str_types.append(k)
            elif is_datetime_type(self.data[k]):
                self.datetime_types.append(k)
            elif is_numeric_type(self.data[k]):
                self.numeric_types.append(k)
            else:
                self.unsupported_types.append(k)
                
        if len(self.unsupported_types)>0:
            print('Failed!')
            raise TypeError('has unsupported types')   ## FIX
            
        if self.orderby:
            if self.orderby not in self.data.columns:
                print('Failed!')
                raise ValueError('The [ orderby ] "{}" column name'
                                 'does not exist in [ data ] dataframe.'.format(self.orderby))

            if self.cross_validation_scheme=='date_rolling_window':
                if self.orderby not in self.datetime_types and self.orderby not in self.date_str_types:
                    print('Failed!')
                    raise TypeError('[ orderby ] argument column in [ data ] dataframe must have either '
                                    'datetime dtype or "YYYY-MM-DD" format if '
                                    '[ cross_validation_scheme ] is specified as "date_rolling_window".')

        if self.groupby:
            if self.groupby not in self.data.columns:
                print('Failed!')
                raise ValueError('The [ groupby ] "{}" column name'
                                 'does not exist in [ data ] dataframe.'.format(self.groupby))

        # convert all date str types to datetime types        
        while len(self.date_str_types)>0:
            # raise warning!
            k = self.date_str_types.pop()
            self.data[k] = pd.to_datetime(self.data[k])
            self.datetime_types.append(k)

        # convert all numeric to float32
        for k in self.numeric_types:
            # raise warning!
            self.data[k] = cast_numeric2float32(self.data[k])

        self.original_colnames = self.data.columns.to_list()
    
    def _validate_hyperparameters(self):
        
        if self.hyperparameters is None:
            return
        
        if isinstance(self.hyperparameters, list):
            self.hyperparameters = {self.dummy_region_uuid: self.hyperparameters} # check fit
            return
        
        if not isinstance(self.hyperparameters, dict):
            print('Failed!')
            raise TypeError('[ hyperparameters ] argument must be either list or dict, '
                            'instead got {}'.format(type(self.hyperparameters)))
            
        if dict_is_nested(self.hyperparameters):
            
            self._check_groupby_presence('hyperparameters')
            self._check_groupby_contents('hyperparameters')
            self._check_orderby_presence('hyperparameters')
                                    
            # check that the dates can be converted to datetime type
            self._check_orderby_contents('hyperparameters')
                
            # check the date range
            # there should be overlaps
            
        else:  # single level dict 
            
            if not self.cross_validation_scheme in ORDERED_CV_SCHEMES:
                # if not, there is no way this is dated keys
                
                self._check_groupby_presence('hyperparameters')
                self._check_groupby_contents('hyperparameters')
                    
            else:
                # if it is ordered, we may have dated key or grouped key
                
                if any([check_date_format(elem) 
                        for elem in list(self.hyperparameters.keys())]):
                    
                    self._check_orderby_presence('hyperparameters')
                    self._check_orderby_contents('hyperparameters')
                        
                    # we need to include the dummy group
                    self.hyperparameters = {self.dummy_region_uuid: self.hyperparameters}
                    
                elif len(set(self.hyperparameters.keys())
                         .intersection(set(self.data[self.groupby].unique()))) > 0:
                    
                    self._check_groupby_presence('hyperparameters')
                    self._check_groupby_contents('hyperparameters')
                    
                    # we need to include the dummy order
                    
                else:
                    print('Failed!')
                    raise ValueError('asdf')
                    
                
    def _check_groupby_presence(self, parameter_name):
        
        if self.groupby is None:
            print('Failed!')
            raise ValueError('If you are specifying per group [ {} ] you '
                             'must supply [ groupby ] argument.'.format(parameter_name))
            
    def _check_groupby_contents(self, parameter_name):
        """Used only when parameter_name parameter is a dictionary type, and against
        the [ data ] dataframe [ groupby ] column contents."""
        
        d = self.__dict__[parameter_name]
        
        if set(d.keys())!=set(self.data[self.groupby].unique()):
            print('Failed!')
            raise ValueError('The [ groupby ] column values of [ data ] dataframe and '
                             'the group keys of [ {} ] dict do not match.'.format(parameter_name))
            
    def _check_orderby_presence(self, parameter_name):
        
        if self.orderby is None:
            print('Failed!')
            raise ValueError('If you are specifying per date [ {} ] you '
                             'must supply [ orderby ] argument.'.format(parameter_name))
            
    def _check_orderby_contents(self, parameter_name):
        
        d = self.__dict__[parameter_name]
        
        if dict_is_nested(d):

            dates = set()
            for k, v in d.items():
                for k_, v_ in v.items():
                    dates.add(k_)
            dates = list(dates)
            
        else:
            
            dates = list(d.keys())
            
        if not all([check_date_format(elem) for elem in dates]):
            print('Failed!')
            raise ValueError('The date keys of [ {} ] failed to be parsed. '
                             'They must be of the form YYYY-MM-DD.'.format(parameter_name))    

            


from ._evaluation_engine.dask_futures import MultiThreadTaskQueue
from ._evaluation_engine.dask_futures import DualClientFuture
from ._evaluation_engine.dask_futures import ClientFuture
from ._evaluation_engine.data_loader import load_local_data
from ._evaluation_engine.data_loader import upload_local_data
from ._evaluation_engine.data_loader import download_local_data
from ._evaluation_engine.data_loader import upload_remote_data
from ._evaluation_engine.data_loader import download_remote_data
from evaluation_framework.utils.objectIO_utils import save_obj
from evaluation_framework.utils.objectIO_utils import load_obj
from evaluation_framework.utils.memmap_utils import write_memmap
from evaluation_framework.utils.memmap_utils import read_memmap
from ._evaluation_engine.cross_validation_split import get_cv_splitter
from .task_graph import TaskGraph

import os
import pandas as pd
import numpy as np
from collections import namedtuple


INSTANCE_TYPES = {
    'm4.large': {'vCPU': 2, 'Mem': 8},
    'm4.xlarge': {'vCPU': 4, 'Mem': 16}, 
    'm4.2xlarge': {'vCPU': 8, 'Mem': 32},
    'm4.4xlarge': {'vCPU': 16, 'Mem': 64},
    'm4.10xlarge': {'vCPU': 40, 'Mem': 160},
    'm4.16xlarge': {'vCPU': 64, 'Mem': 256}, 
    
    'c4.large': {'vCPU': 2, 'Mem': 3.75},
    'c4.xlarge': {'vCPU': 4, 'Mem': 7.5},
    'c4.2xlarge': {'vCPU': 8, 'Mem': 15},
    'c4.4xlarge': {'vCPU': 16, 'Mem': 30},
    'c4.8xlarge': {'vCPU': 36, 'Mem': 60}, 
    
    'r4.large': {'vCPU': 2, 'Mem': 15.25},
    'r4.xlarge': {'vCPU': 4, 'Mem': 30.5},
    'r4.2xlarge': {'vCPU': 8, 'Mem': 61}, 
    'r4.4xlarge': {'vCPU': 16, 'Mem': 122},
    'r4.8xlarge': {'vCPU': 32, 'Mem': 244},
    'r4.16xlarge': {'vCPU': 64, 'Mem': 488}}

DASK_RESOURCE_PARAMETERS = [
    'local_client_n_workers', 
    'local_client_threads_per_worker', 
    'yarn_container_n_workers',
    'yarn_container_worker_vcores', 
    'yarn_container_worker_memory', 
    'n_worker_nodes']

TASK_REQUIRED_KEYWORDS = [
    'memmap_root_dirname',
    'user_configs',
    'preprocess_train_data',
    'model_fit',
    'preprocess_test_data',
    'model_predict',
    'hyperparameters',
    'estimator',
    'feature_names',
    'target_name',
    'evaluate_prediction',
    'orderby',
    'return_predictions',
    'S3_path',
    'memmap_root_S3_object_name',
    'prediction_records_dirname',
    'memmap_root_dirpath',
    'cross_validation_scheme',
    'train_window',
    'test_window']

TaskManager = namedtuple('TaskManager', TASK_REQUIRED_KEYWORDS)


class EvaluationEngine():

    def __init__(self, local_client_n_workers=None, local_client_threads_per_worker=None, 
                 yarn_container_n_workers=None, yarn_container_worker_vcores=None, yarn_container_worker_memory=None,
                 n_worker_nodes=None, use_yarn_cluster=None, use_auto_config=None, instance_type=None, parallelize=True,
                 verbose=True):

        self.parallelize = parallelize
        
        if (local_client_n_workers is not None and 
            local_client_threads_per_worker is not None):
            
            local_client_resources_set = True
            yarn_client_resources_set = False
            
            if (yarn_container_n_workers is not None and 
                yarn_container_worker_vcores is not None and
                yarn_container_worker_memory is not None and
                n_worker_nodes is not None):
                
                yarn_client_resources_set = True
        
        if use_auto_config is None:
            print('\u2714 Set [ use_auto_config ] to True in order to automatically configure Dask resources.\n')
        
        if use_yarn_cluster is None:
            print('\u2714 Set [ use_yarn_cluster ] to True in order to leverage Yarn cluster.\n')
            
        if use_auto_config is None:
            print('You can also manually configure resources by providing arguments for the following parameters:\n\n'
                      '\u25BA {}'.format('  '.join(DASK_RESOURCE_PARAMETERS[0:4])))
            print('\n  ' + '  '.join(DASK_RESOURCE_PARAMETERS[4:]))

        if (use_auto_config is None) or (use_yarn_cluster is None):
            print('\nOptional argument(s):\n\n\u25BA {}'.format('instance_type'))#.join(self.optional_args)))
        
        if use_auto_config:
            
            if use_yarn_cluster:
            
                if (instance_type is None or n_worker_nodes is None):
                    print('\u2714 In order to auto config yarn cluster, please provide the [ instance_type ] and [ n_worker_nodes ].')
                    print('\nEX: instance_type="m4.2xlarge", n_worker_nodes=3 (excluding the master node)')

                    print('\nAvailable [ instance_type ] options: ')
                    print('\n\u25BA {}'.format('  '.join(list(INSTANCE_TYPES.keys())[0:6])))
                    print('\n  ' + '  '.join(list(INSTANCE_TYPES.keys())[6:11]))
                    print('\n  ' + '  '.join(list(INSTANCE_TYPES.keys())[11:]))

                else:
                    num_physical_cores = int(INSTANCE_TYPES[instance_type]['vCPU']/2)
                    off_set = 2 if num_physical_cores>=8 else 1
                    local_client_n_workers = max(1, num_physical_cores - off_set)
                    local_client_threads_per_worker = 2
                    yarn_container_n_workers = num_physical_cores - 1
                    yarn_container_worker_vcores = 2
                    yarn_container_worker_memory = INSTANCE_TYPES[instance_type]['Mem'] - 1.5

            else:
                local_client_n_workers = psutil.cpu_count(logical=False)
                local_client_threads_per_worker = int(psutil.cpu_count(logical=True)/num_physical_cores)
                
        self.use_yarn_cluster = use_yarn_cluster

        self.local_client_n_workers = local_client_n_workers
        self.local_client_threads_per_worker = local_client_threads_per_worker 
        self.yarn_container_n_workers = yarn_container_n_workers
        self.yarn_container_worker_vcores = yarn_container_worker_vcores
        self.yarn_container_worker_memory = yarn_container_worker_memory
        self.n_worker_nodes = n_worker_nodes

        self.verbose = verbose

    def run_evaluation(self, evaluation_manager):

        self.data = evaluation_manager.data

        if self.use_yarn_cluster and evaluation_manager.S3_path is None:
            raise ValueError('if [ use_yarn_cluster ] is set to True, you must provide [ S3_path ] to EvaluationManager object.')

        os.makedirs(evaluation_manager.local_directory_path)
        os.chdir(evaluation_manager.local_directory_path)
        
        print("\u2714 Preparing local data...                ", end="", flush=True)
        memmap_map = load_local_data(evaluation_manager)
        print('Completed!')

        # evaluation_manager is too bulky to travel across network
        self.task_manager = TaskManager(
            **{k: v for k, v in evaluation_manager.__dict__.items() 
            if k in TASK_REQUIRED_KEYWORDS})

        # need condition to open yarn or local!
        if self.use_yarn_cluster:

            print("\u2714 Uploading local data to S3 bucket...   ", end="", flush=True)
            upload_local_data(self.task_manager)
            print('Completed!')

            print("\u2714 Starting Dask client...                ", end="", flush=True)
            self.dask_client = DualClientFuture(local_client_n_workers=self.local_client_n_workers, 
                               local_client_threads_per_worker=self.local_client_threads_per_worker, 
                               yarn_client_n_workers=self.yarn_container_n_workers*self.n_worker_nodes, 
                               yarn_client_worker_vcores=self.yarn_container_worker_vcores, 
                               yarn_client_worker_memory=self.yarn_container_worker_memory)
            print('Completed!')

            print("\u2714 Preparing data on remote workers...    ", end="", flush=True)
            self.dask_client.submit_per_node(download_local_data, self.task_manager)
            print('Completed!')

            num_threads = self.local_client_n_workers + self.yarn_container_n_workers*self.n_worker_nodes

        else:

            print("\u2714 Starting Dask client...                ", end="", flush=True)
            self.dask_client = ClientFuture(local_client_n_workers=self.local_client_n_workers, 
                                   local_client_threads_per_worker=self.local_client_threads_per_worker)
            print('Completed!')
            
            num_threads = self.local_client_n_workers
        
        self.taskq = MultiThreadTaskQueue(num_threads=num_threads)

        if self.verbose:
            print('thread size: {}'.format(num_threads))
                    
        memmap_map_filepath = os.path.join(self.task_manager.memmap_root_dirpath, 'memmap_map')
        memmap_map = load_obj(memmap_map_filepath)
        
        print("\n\u23F3 Starting evaluations...         ", end="", flush=True)
        self.dask_client.get_dashboard_link()
        for group_key in memmap_map['attributes']['sorted_group_keys']:

            if self.task_manager.orderby:

                filepath = os.path.join(memmap_map['root_dirpath'], memmap_map['groups'][group_key]['arrays']['orderby_array']['filepath'])
                dtype = memmap_map['groups'][group_key]['arrays']['orderby_array']['dtype']
                shape = memmap_map['groups'][group_key]['arrays']['orderby_array']['shape']
                group_ordered_array = read_memmap(filepath, dtype, shape)

                cv = get_cv_splitter(
                	self.task_manager.cross_validation_scheme, 
                	self.task_manager.train_window, 
                	self.task_manager.test_window, 
                	group_ordered_array)
                n_splits = cv.get_n_splits()

                task_graph = TaskGraph(self.task_manager, cv)

                for i in range(n_splits):

                    if self.parallelize:
                        self.taskq.put_task(self.dask_client.submit, task_graph.run, group_key, i)
                    else:
                        task_graph.run(group_key, i)

            else:
                pass  # normal cross validations

    def get_evaluation_results(self):

        self.taskq.join()

        res = self.taskq.get_results()
        res_pdf = pd.DataFrame(res, columns=['group_key', 'test_idx', 'eval_result', 'data_count'])
        return res_pdf.sort_values(by=['group_key', 'test_idx']).reset_index(drop=True)

    def get_prediction_results(self, group_key=None):

        self.taskq.join()

        if self.use_yarn_cluster:

            print("\n\u2714 Uploading remote prediction results...   ", end="", flush=True)
            self.dask_client.submit_per_node(upload_remote_data, self.task_manager)
            print('Completed!')

            print("\n\u2714 Downloading remote prediction results... ", end="", flush=True)
            download_remote_data(self.task_manager)
            print('Completed!')

        prediction_dirpath = os.path.join(os.getcwd(), self.task_manager.prediction_records_dirname)
        prediction_filenames = os.listdir(prediction_dirpath)
        prediction_filepaths = [os.path.join(prediction_dirpath, elem) for elem in prediction_filenames]

        prediction_array = np.vstack([np.load(elem) for elem in prediction_filepaths])
        prediction_array = prediction_array[prediction_array[:, 0].argsort()]

        prediction_pdf = pd.DataFrame(prediction_array, columns=['specialEF_float32_UUID', 'specialEF_float32_predictions'])
        prediction_pdf.set_index('specialEF_float32_UUID', inplace=True)
        prediction_pdf = prediction_pdf.reindex(range(0, len(self.data)), fill_value=np.nan)
        self.data['specialEF_float32_predictions'] = prediction_pdf['specialEF_float32_predictions']
        self.data.drop(labels='specialEF_float32_UUID', axis=1, inplace=True)
        return self.data   




from evaluation_framework.utils.data_structure_utils import is_nested_list

import numpy as np


def write_memmap(filepath, dtype, shape, array):

    writable_memmap = np.memmap(filepath, dtype=dtype, mode="w+", shape=shape)
    writable_memmap[:] = array[:]
    del writable_memmap

def read_memmap(filepath, dtype, shape, idx=None):

    readonly_memmap = np.memmap(filepath, dtype=dtype, mode="r", shape=shape)

    if idx is None:
        array = readonly_memmap[:]
    else:
        if is_nested_list(idx):
            array = readonly_memmap[tuple(idx)]
        else:
            array = readonly_memmap[idx]

    del readonly_memmap
    return array




import zipfile
import os
import errno


def unzip_dir(dir_path, dirname):
    
    path, file = os.path.split(dir_path)
    
    new_dir_path = os.path.join(path, dirname)
    
    try:
        os.makedirs(new_dir_path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    
    with zipfile.ZipFile(dir_path, 'r') as f:
        f.extractall(new_dir_path)

    return new_dir_path



from ..default_methods import default_preprocess_train_data
from ..default_methods import default_preprocess_test_data
from ..default_methods import default_model_fit
from ..default_methods import default_model_predict
from ..default_methods import default_evaluate_prediction

import re
import copy


REQUIRED_METHOD_ARGUMENTS = ['preprocess_train_data', 'preprocess_test_data', 'evaluate_prediction']
EVALUATION_OPTIONS = ['mse', 'mae', 'rmse']


class MethodSetter():
	
	def __init__(self):
		
		self.num_types_needed = None
		self.missing_keys = dict()
		self.missing_keys['datetime_types'] = []
		self.missing_keys['str_types'] = []
		
	def set_methods(self, config_setter=None, 
					preprocess_train_data="", preprocess_test_data="",
					model_fit=None, model_predict=None,
					store_prediction=None, evaluate_prediction="", **kwargs):

		if config_setter.groupby is None:
			config_setter.groupby = config_setter.dummy_region_uuid
		
		self.config_setter = config_setter
		self.preprocess_train_data = preprocess_train_data
		self.preprocess_test_data = preprocess_test_data
		
		self.model_fit = model_fit
		self.model_predict = model_predict
		self.store_prediction = store_prediction
		self.evaluate_prediction = evaluate_prediction
		
		print("\u2714 Checking methods requirements...    ", end="", flush=True)
		if not self.passed_arguments_requirements():
			return False
		else:
			print('Passed!')

		print("\u2714 Checking methods validity...        ", end="", flush=True)
		if not self.passed_arguments_validity():
			return False
		else:
			print('Passed!')
		
		return True
	
	def passed_arguments_requirements(self):

		# if self.config_setter is None:
		# 	print('You must')
		# 	return False

		requirements = copy.copy(REQUIRED_METHOD_ARGUMENTS)
		
		if self.preprocess_train_data!="":
			requirements.remove('preprocess_train_data')
		if self.preprocess_test_data!="":
			requirements.remove('preprocess_test_data')
		if self.evaluate_prediction!="":
			requirements.remove('evaluate_prediction')

		if len(requirements)>0:
			print('Failed!')
			print('\nRequired method argument(s):\n\n\u25BA {}'.format('  '.join(requirements)))

			if 'evaluate_prediction' in requirements:
				print('\nAvailable predefined [ evaluate_prediction ] options:\n\n'
                  '\u25BA {}'.format('  '.join(EVALUATION_OPTIONS)))

			print('\nIf you pass in None, default methods will be used instead.')

			return False
		else:
			return True

	def passed_arguments_validity(self):
		
		# need to check they passed in arguments correctly
		# need to check that they do not require configs if nothing is passed in
		
		
		self._validate_preprocess_train_data_method()

		self._validate_preprocess_test_data()
		self._validate_model_fit()
		self._validate_model_predict()
		# self._validate_store_prediction()
		self._validate_evaluate_prediction()

		self.num_types_needed = self._num_types_needed()

		return True
		
		
	def _validate_preprocess_train_data_method(self):
		
		if self.preprocess_train_data is None:
			
			self.preprocess_train_data = default_preprocess_train_data
			return
	
	def _validate_preprocess_test_data(self):
		
		if self.preprocess_test_data is None:
			
			self.preprocess_test_data = default_preprocess_test_data
			return
	
	def _validate_model_fit(self):
		
		if self.model_fit is None:
			
			self.model_fit = default_model_fit
			return
	
	def _validate_model_predict(self):
		
		if self.model_predict is None:
			
			self.model_predict = default_model_predict
			return
	
	def _validate_store_prediction(self):
		if self.store_prediction is None:
			
			self.store_prediction = default_store_prediction
			return
	
	def _validate_evaluate_prediction(self):
		if self.evaluate_prediction is None:
			
			self.evaluate_prediction = default_evaluate_prediction
			return
		
	def _num_types_needed(self):

		self.sample_train_pdf, self.sample_test_pdf = self._get_sample_pdf(self.config_setter)

		included_colnames = copy.copy(self.config_setter.numeric_types)
		train_missing_keys, re = self.key_error_catcher(
			self.preprocess_train_data, 
			self.sample_train_pdf[included_colnames], 
			self.config_setter.user_configs)
		test_missing_keys, re = self.key_error_catcher(
			self.preprocess_test_data, 
			self.sample_test_pdf[included_colnames], 
			re,
			self.config_setter.user_configs)
		missing_keys = train_missing_keys + test_missing_keys
		
		if len(missing_keys)==0:
			return 1

		self.missing_keys['datetime_types'] = copy.copy(missing_keys)
		
		if len(missing_keys)>0:
			
			included_colnames += self.config_setter.datetime_types
			train_missing_keys, re = self.key_error_catcher(
				self.preprocess_train_data, 
				self.sample_train_pdf[included_colnames], 
				self.config_setter.user_configs)
			test_missing_keys, re = self.key_error_catcher(
				self.preprocess_test_data, 
				self.sample_test_pdf[included_colnames], 
				re, 
				self.config_setter.user_configs)
			missing_keys = train_missing_keys + test_missing_keys
			
		if len(missing_keys)==0:
			return 2

		self.missing_keys['datetime_types'] = list(set(self.missing_keys['datetime_types']) - set(missing_keys))
		self.missing_keys['str_types'] = copy.copy(missing_keys)
			
		if len(missing_keys)>0:
			
			included_colnames += self.config_setter.str_types
			train_missing_keys, re = self.key_error_catcher(
				self.preprocess_train_data, 
				self.sample_train_pdf[included_colnames], 
				self.config_setter.user_configs)
			test_missing_keys, re = self.key_error_catcher(
				self.preprocess_test_data, 
				self.sample_test_pdf[included_colnames], 
				re, 
				self.config_setter.user_configs)
			missing_keys = train_missing_keys + test_missing_keys
			
		if len(missing_keys)==0:
			return 3
		
		else:
			raise ValueError('not included! {}'.format(missing_keys))

	def _get_sample_pdf(self, config_setter):

		# do the groupby size ordering here! and get the smallest one!
		# by the str! that we can use it later

		if config_setter.groupby:
			
			sample_group = config_setter.data[config_setter.groupby].iloc[0]
			
			sample_pdf = config_setter.data[config_setter.data[config_setter.groupby]==sample_group]
			
			n = min(len(sample_pdf), 1000)
			train_n = int(n * 0.75)
			
			sample_train_pdf = sample_pdf.iloc[0:train_n]
			sample_test_pdf = sample_pdf.iloc[train_n:]
			
		else:

			n = min(len(config_setter.data), 1000)
			train_n = int(n * 0.75)

			sample_pdf = config_setter.data.iloc[0:n]
			sample_train_pdf = sample_pdf.iloc[0:train_n]
			sample_test_pdf = sample_pdf.iloc[train_n:]

		return sample_train_pdf, sample_test_pdf

	def key_error_catcher(self, f, *args, **kwargs):
	
		try:
			re = f(*args, **kwargs)
			return [], re

		except KeyError as e:

			key = e.args[0]

			if key in self.config_setter.original_colnames:
				return [key], None

			key = eval(re.search('(\[\'.*\'\])', key, re.IGNORECASE).group(1))

			if set(key) < set(self.config_setter.original_colnames):
				return key, None
			
			return key, None
			



from setuptools import setup, find_packages


import evaluation_framework

VERSION = evaluation_framework.__version__

# with open("README.rst", "r") as fh:
#     long_description = fh.read()

setup(
	name="ml-evaluation-framework", 
	version=VERSION,
	author="Jay Kim",
	description="",
	# long_description=long_description,
	# long_description_content_type="text/x-rst",
	url=None,
	license="DSB 3-clause",
	packages=find_packages(),
	# install_requires=["graphviz>=0.13.2"]
	)

# python setup.py bdist_wheel
# scp dist/evaluation_framework-0.0b0-py3-none-any.whl jkim2@10.174.62.128:/mnt/grubhub-gdp-delivery-data-assets-dev/notebooks/jkim2/model_experiments/evaluation_framework


from collections import defaultdict


def dict_is_nested(d):
    
    return any(isinstance(i,dict) for i in d.values())

def get_merged_list_from_dict_list_values(d):
    
    s = set()
    
    for k, v in d.items():
        
        s = s.union(v)
    
    return list(s)

def is_nested_list(l):
    
    if not isinstance(l, list):
        return False
    else:
        return any(isinstance(i, list) for i in l)
        


import pickle


def save_obj(obj, filepath):
    with open(filepath, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        
def load_obj(filepath):
    with open(filepath, 'rb') as f:
        return pickle.load(f)



from abc import ABCMeta, abstractmethod
import numpy as np

from sklearn.model_selection._split import *
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples

__all__ = ['BaseCrossValidator',
           'KFold',
           'GroupKFold',
           'LeaveOneGroupOut',
           'LeaveOneOut',
           'LeavePGroupsOut',
           'LeavePOut',
           'RepeatedStratifiedKFold',
           'RepeatedKFold',
           'ShuffleSplit',
           'GroupShuffleSplit',
           'StratifiedKFold',
           'StratifiedShuffleSplit',
           'PredefinedSplit',
           'train_test_split',
           'check_cv',
           'BaseRollingWindowSplit',
           'DateRollingWindowSplit']


def get_cv_splitter(cross_validation_scheme, *args, **kwargs):

    if cross_validation_scheme=='date_rolling_window':
        return DateRollingWindowSplit(*args, **kwargs)
    elif cross_validation_scheme=='k_fold':
        return KFold(*args, **kwargs)


class BaseRollingWindowSplit(metaclass=ABCMeta):
    
    def __init__(self, orderby=None, random_state=None):
        self.orderby = orderby
        self.random_state = random_state
    
    def split(self, X, y=None, groups=None):
        X, y, groups = indexable(X, y, groups)
        for train, test in self._iter_indices(X, y, groups):
            yield train, test
    
    @abstractmethod
    def _iter_indices(self, X, y=None, groups=None):
        """Generate (train, test) indices"""
    
    @abstractmethod
    def get_n_splits(self, X=None, y=None, groups=None):
        """Get the total number of splits"""
    
    def __repr__(self):
        return _build_repr(self)
        

class DateRollingWindowSplit(BaseRollingWindowSplit):
    
    def __init__(self, num_train_days, num_test_days, orderby,
                 random_state=None):
        super().__init__(random_state=random_state, orderby=orderby)
        self.num_train_days = num_train_days
        self.num_test_days = num_test_days
    
    def _iter_indices(self, X, y, groups):
        
        n_samples = _num_samples(X)
        indices = np.arange(n_samples)
        head_date_idx = 0
        last_date_idx = self.orderby.max()

        while(head_date_idx + self.num_train_days <= last_date_idx):

            train = indices[(head_date_idx <= self.orderby) & 
                            (self.orderby < head_date_idx + self.num_train_days)]
            test = indices[(head_date_idx + self.num_train_days <= self.orderby) & 
                           (self.orderby < head_date_idx + self.num_train_days + self.num_test_days)]

            head_date_idx += self.num_test_days
            
            yield train, test
            
    def get_n_splits(self, X=None, y=None, groups=None):
        
        head_date_idx = 0
        last_date_idx = self.orderby.max()
        train_data_cnt = 0
        
        while(head_date_idx + self.num_train_days <= last_date_idx):

            head_date_idx += self.num_test_days
            train_data_cnt += 1
            
        return train_data_cnt

    def split(self, X, y=None, groups=None):
        return super().split(X, y, groups)


def _build_repr(self):
    # XXX This is copied from BaseEstimator's get_params
    cls = self.__class__
    init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
    # Ignore varargs, kw and default values and pop self
    init_signature = signature(init)
    # Consider the constructor parameters excluding 'self'
    if init is object.__init__:
        args = []
    else:
        args = sorted([p.name for p in init_signature.parameters.values()
                       if p.name != 'self' and p.kind != p.VAR_KEYWORD])
    class_name = self.__class__.__name__
    params = dict()
    for key in args:
        # We need deprecation warnings to always be on in order to
        # catch deprecated param values.
        # This is set in utils/__init__.py but it gets overwritten
        # when running under python3 somehow.
        warnings.simplefilter("always", FutureWarning)
        try:
            with warnings.catch_warnings(record=True) as w:
                value = getattr(self, key, None)
                if value is None and hasattr(self, 'cvargs'):
                    value = self.cvargs.get(key, None)
            if len(w) and w[0].category == FutureWarning:
                # if the parameter is deprecated, don't show it
                continue
        finally:
            warnings.filters.pop(0)
        params[key] = value

    return '%s(%s)' % (class_name, _pprint(params, offset=len(class_name)))

